#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module encapsulates all shell callable functions of nidaba.
"""

from __future__ import absolute_import, print_function, unicode_literals

from nidaba import Batch, storage
from nidaba.config import nidaba_cfg
from nidaba import celery
from pprint import pprint
from inspect import getcallargs, getdoc

import argparse
import re
import uuid
import shutil
import os.path
import sys
import click
import pkg_resources

@click.group(epilog='This nidaba may or may not have Super Cow Powers')
@click.version_option()
def main():
    """
    Sends jobs to nidaba and retrieves their status.
    """


def int_float_or_str(s):
    """
    A small helper function intended to coerce an input string to types in the
    order int -> float -> unicode -> input.

    Args:
        s (unicode):

    Returns:
        int or float or unicode or original input type: Input variable coerced
        to the highest data type in the ordering.
    """

    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            try:
                return unicode(s)
            except UnicodeDecodeError:
                return s

def get_prefix_tasks(prefix=''):
    """
    Returns a list of all registered tasks (including in plugins) that start
    with a certain prefix, i.e. are 
    """
    t = celery.app.tasks
    return [k.split('.')[-1] for k in t.iterkeys() if k.startswith('nidaba.' +
            prefix)]


def help_tasks(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    t = celery.app.tasks
    hidden = ['util']
    last = u''
    docs = u''
    for k in sorted(t.keys()):
        hier = k.split('.')
        if hier[0] == 'nidaba':
            if hier[1] in hidden:
                continue
            elif hier[1] != last:
                docs += '{}\n{}\n\n'.format(hier[1].title(), len(hier[1]) *
                                            '-')
                last = hier[1]
            docs += '{}\n{}\n\n{}'.format(hier[-1], len(hier[-1]) * '~',
                                          getdoc(t[k].run).decode('utf-8').partition('Returns:')[0])
    click.echo_via_pager(docs)
    ctx.exit()

def validate_definition(ctx, param, value):
    """
    Validates all task definitions of a group and returns them as a list.
    """
    definitions = []
    for alg in value:
        (task, _, params) = alg.partition(u':')
        task = param.human_readable_name + '.' + task
        if 'nidaba.' + task not in celery.app.tasks:
            raise click.BadParameter('Unknown task ' + task)
        configurations = []
        for conf in params.split(u';'):
            args = []
            kwargs = {}
            for arg in re.split(r'(?<!\\):', conf):
                # treat as kwarg
                if '=' in arg:
                    k, v = arg.split('=')
                    kwargs[k] = int_float_or_str(v)
                # treat as arg as long as no kwargs were defined
                elif not kwargs and len(arg):
                    args.append(int_float_or_str(arg))
                elif not len(arg):
                    continue
                else:
                    raise click.BadParameter('Positional argument after '
                                             'keyword argument')
            try:
                fun = celery.app.tasks['nidaba.' + task]
                # fill in common kwargs (doc/method)
                args.insert(0, '')
                args.insert(0, '')
                kwargs = getcallargs(fun.run, *args, **kwargs)
                kwargs.pop('doc')
                kwargs.pop('method')
            except TypeError as e:
                raise click.BadParameter(e.message)
            configurations.append(kwargs)
        definitions.append([task, configurations])
    return definitions

@main.command()
@click.option('--binarize', '-b', multiple=True, callback=validate_definition,
              help='A configuration for a single binarization algorithm in '
              'the format algorithm:param1,param2;param1,param2;...')
@click.option('--ocr', '-o', multiple=True, callback=validate_definition, 
              help='A configuration for a single OCR engine in the format '
              'engine:param1,param2;param1,param2;...')
@click.option('--stats', '-s', multiple=True, callback=validate_definition,
              help='A configuration for a single post-OCR measure in the '
              'format measure:param1,param2;param1;param2...')
@click.option('--willitblend', 'blend',  default=False, help='Blend all '
              'output files into a single hOCR document.', is_flag=True)
@click.option('--grayscale', default=False, help='Skip grayscale '
              'conversion using the ITU-R 601-2 luma transform if the input '
              'documents are already in grayscale.', is_flag=True)
@click.option('--jobid', default=uuid.uuid4(), type=str, help='Force a job '
              'identifier. This may or may not be an UUID but it has to be an '
              'unused identifer.')
@click.option('--help-tasks', is_eager=True, is_flag=True, callback=help_tasks,
              help='Accesses the documentation of all tasks contained in '
              'nidaba itself and in configured plugins.')
@click.argument('files', type=click.Path(exists=True), nargs=-1, required=True)
def batch(files, binarize, ocr, stats, blend, grayscale, jobid, help_tasks):
    """
    Add a new job to the pipeline.
    """
    id = unicode(jobid)
    batch = Batch(id)
    click.echo(u'Preparing filestore\t\t[', nl=False),
    if storage.prepare_filestore(id) is None:
        click.secho(u'\u2717', fg='red', nl=False)
        click.echo(']')
        exit()
    for doc in files:
        shutil.copy2(doc, storage.get_abs_path(id, os.path.basename(doc)))
        batch.add_document((id, os.path.basename(doc)))
    click.secho(u'\u2713', fg='green', nl=False)
    click.echo(']')
    click.echo(u'Building batch\t\t\t[', nl=False)
    batch.add_step()
    if not grayscale:
        batch.add_tick()
        batch.add_task('img.rgb_to_gray')
    if binarize:
        batch.add_tick()
        for alg in binarize:
            for kwargs in alg[1]:
                batch.add_task(alg[0], **kwargs)
    if ocr:
        batch.add_tick()
        for alg in ocr:
            for kwargs in alg[1]:
                batch.add_task(alg[0], **kwargs)
    if stats:
        batch.add_tick()
        for alg in stats:
            for kwargs in alg[1]:
                batch.add_task(alg[0], **kwargs)
    if blend:
        batch.add_step()
        batch.add_tick()
        batch.add_task('postprocessing.blend_hocr')
    batch.run()
    click.secho(u'\u2713', fg='green', nl=False)
    click.echo(']')
    click.echo(id)


@main.command()
def worker():
    """
    Starts a celery worker.
    """
    celery.app.worker_main(argv=sys.argv[:1])


@main.command()
def config():
    """
    Displays the current configuration.
    """
    pprint(nidaba_cfg)


@main.command()
@click.argument('job_id', nargs=1, type=str)
def status(job_id):
    """
    Diplays the status and results of jobs.
    """
    batch = Batch(job_id)
    state = batch.get_state()
    print(state)
    if state == 'SUCCESS':
        ret = batch.get_results()
        if ret is None:
            print('Something somewhere went wrong.')
            print('Please contact your friendly nidaba support technician.')
        else:
            for doc in ret:
                click.echo(doc['root'][1].encode('utf-8') + u' \u2192 ' + 
                      storage.get_abs_path(*doc['doc']).encode('utf-8'))
    elif state == 'FAILURE':
        ret = batch.get_errors()
        if ret is None:
            print('Something somewhere went wrong.')
        else:
            for fun in ret:
                print(fun[0]['method'].encode('utf-8'),
                      'failed while operating on',
                      fun[0]['doc'][1].encode('utf-8'),
                      'which is based on',
                      fun[1]['root'][1].encode('utf-8'))



# -*- coding: utf-8 -*-
"""
nidaba.algorithms.string
~~~~~~~~~~~~~~~~~~~~~~~~

Implementation of various algorithms operating on strings and unicode objects,
e.g. alignment, edit distances, and symmetric deletion searches.
"""

from __future__ import division, absolute_import

import codecs
import numpy
import operator
import unicodedata
import itertools
import mmap
import math

from nidaba.nidabaexceptions import (NidabaUnibarrierException,
                                     NidabaAlgorithmException)

# ----------------------------------------------------------------------
# String and alignment algorithms --------------------------------------
# ----------------------------------------------------------------------


def unibarrier(func):
    """
    A decorator function; used to ensure that no str objects can be
    passed as either args or kwargs.
    """

    def unishielded(*args, **kwargs):
        for arg in args:
            if isinstance(arg, str):
                raise NidabaUnibarrierException(message='%s was a string!' %
                                                arg)
        for key, val in kwargs.iteritems():
            if isinstance(val, str):
                raise NidabaUnibarrierException(message='%s was a string!' %
                                                val)
        return func(*args, **kwargs)
    return unishielded


def sanitize(string, encoding=u'utf-8', normalization=u'NFD'):
    """
    Strip leading and trailing whitespace, convert to NFD. If the passed
    string is a str rather than an unicode, decode it with the specified
    encoding.
    """
    if isinstance(string, str):
        string = string.decode(encoding)

    return unicodedata.normalize(normalization, string.strip())


@unibarrier
def strings_by_deletion(unistr, dels):
    """
    Compute the unique strings which can be formed from a string by
    deleting the specified number of characters from it. The results
    are sorted in ascending order.
    """
    new_words = set()
    for comb in itertools.combinations(range(len(unistr)), dels):
        new_words.add(u''.join((c for i, c in enumerate(unistr) if i not in
                                comb)))
    return sorted(list(new_words))


@unibarrier
def sym_suggest(ustr, dic, delete_dic, depth, ret_count=0):
    """
    Return a list of "spelling" corrections using a symmetric deletion search.
    Dic is a set of correct words. Delete_dic is of the form
    {edit_term:[(candidate1, edit_distance), (candidate2, edit_distance),
    ...]}.
    """
    suggestions = set()
    dels = strings_by_deletion(ustr, depth)
    if ustr in dic:
        suggestions.add(ustr)

    if ustr in delete_dic:  # ustr is missing characters.
        suggestions = suggestions.union(set(delete_dic[ustr]))
    for s in dels:
        if s in dic:    # ustr has extra characters.
            suggestions.add(s)
        if s in delete_dic:  # ustr has substitutions of twiddles
            suggestions = suggestions.union(set(delete_dic[s]))

    return list(suggestions if ret_count <= 0 else suggestions[:ret_count])


@unibarrier
def parse_del_dict_entry(entry):
    return [] if entry is None else [word.strip()
                                     for word in entry.split(u' ')]


@unibarrier
def suggestions(ustr, sugs, freq=None):
    """
    Call mapped_sym_suggest, and return the suggestions as a
    sorted list. Python's built in sort is stable, so we can
    simply sort repeatedly, from least important aspect to
    most important.
    """
    sugs = sorted(sugs)  # Alphabetic sort
    if freq is not None:
        sugs = sorted(sorted(sugs), key=lambda x: freq[x])  # By frequency
    # By edit distance
    sugs = sorted(sugs, key=lambda x: edit_distance(ustr, x))

    return sugs


@unibarrier
def mapped_sym_suggest(ustr, del_dic_path, dic, depth, ret_count=0):
    """
    Generate a list of spelling suggestions using the memory mapped
    dictionary search/symmetric delete algorithm. Return only
    suggestions at the specified depth, not up to and including that
    depth.
    """
    deletes = set()
    inserts = set()
    int_and_dels = set()
    subs = set()

    dels = strings_by_deletion(ustr, depth)
    ustr_entry = mmap_bin_search(ustr, del_dic_path)
    word_for_ustr = parse_del_dict_entry(ustr_entry)
    if word_for_ustr is not None:
        # get the words reachable by adding to ustr.
        inserts = set(w for w in word_for_ustr)
    for s in dels:
        if s in dic:
            deletes.add(s)  # Add a word reachable by deleting from ustr.

        line_for_s = parse_del_dict_entry(mmap_bin_search(s, del_dic_path))
        if line_for_s is not None:
            # Get the words reachable by deleting from originals, adding to
            # them. Note that this is NOT the same as 'Levenshtein'
            # substitution.
            for sug in line_for_s:
                distance = edit_distance(sug, ustr)
                if distance == depth:
                    subs.add(sug)
                elif distance > depth:
                    int_and_dels.add(sug)

    return {u'dels': deletes, u'ins': inserts,
            u'subs': subs, u'ins+dels': int_and_dels}


def prev_newline(mm, line_buffer_size=100):
    """
    Return the pointer position immediately after the closest left hand
    newline, or to the beginning of the file if no such newlines exist.
    """
    # mm.seek(mm.tell - line_buffer_size)
    # TODO this fails on a line greater than line_buffer_size in length
    return mm.rfind(u'\n', mm.tell() - line_buffer_size, mm.tell()) + 1


@unibarrier
def compare_strings(u1, u2):
    if u1 == u2:
        return 0
    elif u1 < u2:
        return 1
    else:
        return -1


@unibarrier
def todec(ustr):
    uthex = u''
    for cp in ustr:
        uthex += u'<' + cp + u'>' + u',' + unicode(ord(cp)) + u':'
    return uthex.encode('utf-8')


@unibarrier
def truestring(unicode):
    out = u'<' + u':'.join([u for u in unicode]) + u'>'
    return out

# ----------------------------------------------------------------------
# Binary search dictionary entry parsers -------------------------------
# ----------------------------------------------------------------------


@unibarrier
def key_for_del_dict_entry(entry):
    """
    Parse a line from a symmetric delete dictionary.
    Returns a tuple of the form (key, list of values).
    """
    key, val = entry.split(u'\t')
    return (key, val.strip())


@unibarrier
def key_for_single_word(entry):
    """
    Parse a line from a simple "one word per line" dictionary.
    """
    cleanword = entry.strip()
    return (cleanword, cleanword)

# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------


# TODO Implement doubling-length backward search to make line_buffer_size
# irrelevant.

@unibarrier
def mmap_bin_search(ustr, dictionary_path,
                    entryparser_fn=key_for_del_dict_entry,
                    line_buffer_size=200):
    """
    Perform a binary search on a memory mapped dictionary file, and
    return the parsed entry, or None if the specified entry cannot be
    found. This function assumes that the dictionary is properly
    formatted and well-formed, otherwise the behavior is undefined.
    Line buffer must not be shorter than the longest line in the
    dictionary. Entries may be any strings which do not contain newlines
    (newlines delimint entries); the entryparser_fn should be of the
    form fn_name(unicodestr), decorated with @unibarrier and return a
    tuple of the form (keytosort by, val). By default, it uses the
    function for parsing symmetric deletion dictionary entries.
    The line_buffer_size argument must be >= the longest line in the
    dictionary, or behavior is undefined.
    """

    def current_entry(mm):
        start = mm.tell()
        rawline = mm.readline()
        mm.seek(start)
        return entryparser_fn(rawline.decode(u'utf-8'))

    with codecs.open(dictionary_path, 'r+b') as f:
        # memory-map the file, size 0 means whole file
        mm = mmap.mmap(f.fileno(), 0)
        imin = 0
        imax = mm.size()
        count = 0
        while True:
            mid = imin + int(math.floor((imax - imin) / 2))
            mm.seek(mid)
            mm.seek(prev_newline(mm))
            key, entry = current_entry(mm)

            if key == ustr:
                return entry
            elif key < ustr:
                imin = mid + 1
            else:
                imax = mid - 1

            count += 1
            if imin >= imax:
                break
        return None

# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------


def initmatrix(rows, columns, defaultval=0):
    """Initializes a 2d list to the desired dimensions."""
    return [[defaultval for j in xrange(columns)] for i in xrange(rows)]


def mr(matrix):
    """Returns a string rep of a 2d list as a matrix. Useful for
    debugging."""
    if matrix == []:
        return '[]'
    string = '['
    count = 0
    for row in matrix:
        if count >= 1:
            string += ' '
        string += str(row) + '\n'
        count += 1
    string = string[:-1]
    string += ']'
    return string


def native_backtrace(matrix, start=None):
    """Trace edit steps backward to find an edit sequence for an
    alignment. Starts at the provided 'start' index, or in the i,j'th
    index if none is provided. The backtrace always ends
    at the index 0,0."""
    i, j = start if start is not None else (
        len(matrix) - 1, len(matrix[0]) - 1)
    key = {'i': (0, -1), 'd': (-1, 0), 'm': (-1, -1), 's': (-1, -1)}
    path = []

    while matrix[i][j] != '':
        path.insert(0, matrix[i][j])
        i, j = tuple(map(operator.add, (i, j), key[matrix[i][j]]))

    return path


def native_align(str1, str2, substitutionscore=1, insertscore=1, deletescore=1,
                 charmatrix={}):
    """Calculate the edit distance of two strings, then backtrace to
    find a valid edit sequence."""
    matrix, steps = native_full_edit_distance(str1, str2,
                                              substitutionscore=substitutionscore,
                                              insertscore=insertscore,
                                              deletescore=deletescore,
                                              charmatrix=charmatrix)
    return native_backtrace(steps)


def native_semi_global_align(shortseq, longseq, substitutionscore=1,
                             insertscore=1, deletescore=1, charmatrix={}):
    if len(shortseq) > len(longseq):
        raise NidabaAlgorithmException('shortseq must be <= longseq in\
                                       length!')

    matrix, steps = native_full_edit_distance(shortseq, longseq,
                                              substitutionscore=substitutionscore,
                                              insertscore=insertscore,
                                              deletescore=deletescore,
                                              charmatrix=charmatrix,
                                              alignment_type='semi-global')
    return native_backtrace(
        steps, start=(len(matrix) - 1, matrix[-1].index(min(matrix[-1]))))


def native_global_matrix(str1, str2, substitutionscore, insertscore,
                         deletescore, charmatrix):
    """An initial matrix for a global sequence alignment."""
    matrix = initmatrix(len(str1) + 1, len(str2) + 1)
    for i in xrange(1, len(matrix)):
        matrix[i][0] = i * charmatrix.get(('', str1[i - 1]), deletescore)
    for j in xrange(1, len(matrix[0])):
        matrix[0][j] = j * charmatrix.get((str2[j - 1], ''), insertscore)
    return matrix


def native_semi_global_matrix(str1, str2, substitutionscore, insertscore,
                              deletescore, charmatrix):
    matrix = initmatrix(len(str1) + 1, len(str2) + 1)
    for i in xrange(1, len(matrix)):
        matrix[i][0] = i * charmatrix.get(('', str1[i - 1]), deletescore)
    return matrix


def native_full_edit_distance(str1, str2, substitutionscore=1, insertscore=1,
                              deletescore=1, charmatrix={},
                              alignment_type='global'):

    matrix, steps = full_edit_distance(str1, str2,
                                       substitutionscore=substitutionscore,
                                       insertscore=insertscore,
                                       deletescore=deletescore,
                                       charmatrix=charmatrix,
                                       alignment_type=alignment_type)
    return matrix, steps


def edit_distance(str1, str2, substitutionscore=1, insertscore=1,
                  deletescore=1, charmatrix={},
                  alignment_type='global'):
    m = native_full_edit_distance(str1, str2,
                                  substitutionscore=substitutionscore,
                                  insertscore=insertscore,
                                  deletescore=deletescore,
                                  charmatrix=charmatrix,
                                  alignment_type=alignment_type)[0]
    return m[-1][-1]


def full_edit_distance(str1, str2, substitutionscore=1, insertscore=1,
                       deletescore=1, ins_func=None, iargs=[], ikwargs={},
                       del_func=None, dargs=[], dkwargs={}, sub_func=None,
                       sargs=[], skwargs={}, charmatrix={},
                       alignment_type='global'):
    """
    A version of the modified Wagner-Fischer algorithm that accepts user
    defined scoreing functions. These functions should be of the form
    fname(token1, token2, args*, kwargs**) and return an integer
    >= 0. A return value of 0 indicates an optimality. The larger, the
    integer, the worse the score. Charmatrix is here only used to
    calulate default delete and insert scores for the initial matrix.
    """

    def dscore(c1, c2, default):
        return charmatrix.get((c1, c2), default)
    if ins_func is None:
        ins_func = dscore
        iargs = [insertscore]
    if del_func is None:
        del_func = dscore
        dargs = [deletescore]
    if sub_func is None:
        sub_func = dscore
        sargs = [substitutionscore]

    types = {'global': native_global_matrix,
             'semi-global': native_semi_global_matrix}
    matrix = types[alignment_type](str1, str2, substitutionscore, insertscore,
                                   deletescore, charmatrix)

    steps = initmatrix(len(str1) + 1, len(str2) + 1, defaultval='')
    steps[0][1:] = list('i' * (len(str2)))
    for idx in xrange(1, len(matrix)):
        steps[idx][0] = 'd'
    for i in xrange(1, len(matrix)):
        for j in xrange(1, len(matrix[0])):
            c1 = str1[i - 1]
            c2 = str2[j - 1]

            scores = (('s', matrix[i - 1][j - 1] + sub_func(c1, c2, *sargs,
                                                            **skwargs)),
                      ('i', matrix[i][j - 1] + ins_func(c1, c2, *iargs,
                                                        **ikwargs)),
                      ('d', matrix[i - 1][j] + del_func(c1, c2, *dargs,
                                                        **dkwargs)))
            if str1[i - 1] == str2[j - 1]:
                matrix[i][j] = matrix[i - 1][j - 1]
                steps[i][j] = 'm'
            else:
                bestoption = min(scores, key=lambda x: x[1])
                matrix[i][j] = bestoption[1]
                steps[i][j] = bestoption[0]

    return matrix, steps

# ----------------------------------------------------------------------
# String and alignment algorithms (numpy versions) ---------------------
# ----------------------------------------------------------------------


def np_backtrace(matrix, start=None):
    """Trace edit steps backward to find an edit sequence for an
    alignment. Starts at the provided 'start' index, or in the i,j'th
    index if none is provided. The backtrace always ends
    at the index 0,0."""

    i, j = start if start is not None else (
        matrix.shape[0] - 1, matrix.shape[1] - 1)
    key = {'i': (0, -1), 'd': (-1, 0), 'm': (-1, -1), 's': (-1, -1)}
    path = []

    while matrix[i, j] != '':
        path.insert(0, matrix[i, j])
        i, j = tuple(map(operator.add, (i, j), key[matrix[i, j]]))

    return path


def np_align(str1, str2, substitutionscore=1, insertscore=1, deletescore=1,
             charmatrix={}):
    """Calculate the edit distance of two strings, then backtrace to
    find a valid edit sequence."""
    matrix, steps = np_full_edit_distance(str1, str2,
                                          substitutionscore=substitutionscore,
                                          insertscore=insertscore,
                                          deletescore=deletescore,
                                          charmatrix=charmatrix)
    return np_backtrace(steps)


def np_semi_global_align(shortseq, longseq, substitutionscore=1, insertscore=1,
                         deletescore=1, charmatrix={}):
    """Find a semi-global alignment between two strings."""
    if len(shortseq) > len(longseq):
        raise NidabaAlgorithmException('shortseq must be <= longseq in\
                                       length!')

    matrix, steps = np_full_edit_distance(shortseq, longseq,
                                          substitutionscore=substitutionscore,
                                          insertscore=insertscore,
                                          deletescore=deletescore,
                                          charmatrix=charmatrix,
                                          alignment_type='semi-global')
    back = np_backtrace(steps, start=(matrix.shape[0] - 1,
                                      numpy.argmin(matrix[-1:])))
    return back


def np_global_matrix(str1, str2, substitutionscore, insertscore, deletescore,
                     charmatrix):
    """An initial matrix for a global sequence alignment."""
    matrix = numpy.empty(shape=(str1.size + 1, str2.size + 1))
    matrix[0, 0] = 0
    for i in xrange(1, matrix.shape[0]):
        matrix[i, 0] = i * charmatrix.get(('', str1[i - 1]), deletescore)
    for j in xrange(1, matrix.shape[1]):
        matrix[0, j] = j * charmatrix.get((str2[j - 1], ''), insertscore)
    return matrix


def np_semi_global_matrix(str1, str2, substitutionscore, insertscore,
                          deletescore, charmatrix):
    """An initial matrix for a semi-global sequence alignment."""
    matrix = numpy.zeros(shape=(str1.size + 1, str2.size + 1))

    # We assume that str1 >= str2.
    # This is guaranteed by the semi_global_align function.
    for i in xrange(1, matrix.shape[0]):
        matrix[i, 0] = i * charmatrix.get(('', str1[i - 1]), deletescore)
    return matrix


def np_full_edit_distance(str1, str2, substitutionscore=1, insertscore=1,
                          deletescore=1, charmatrix={},
                          alignment_type='global'):
    """A modified implenmentation of the Wagner-Fischer algorithm using
    numpy. Unlike the minimal and optimized version in the
    "edit_distance" function, this returns the entire scoring matrix,
    and an operation matrix for backtracing and reconstructing the
    edit operations. This should be used when an alignment is desired,
    not only the edit distance."""

    str1 = numpy.array(tuple(str1))
    str2 = numpy.array(tuple(str2))

    types = {'global': np_global_matrix, 'semi-global': np_semi_global_matrix}
    matrix = types[alignment_type](str1, str2, substitutionscore,
                                   insertscore, deletescore, charmatrix)

    steps = numpy.empty(
        shape=(str1.size + 1, str2.size + 1), dtype=numpy.object)
    steps[1:, 0] = 'd'
    steps[0, 1:] = 'i'
    steps[0, 0] = ''
    for i in xrange(1, matrix.shape[0]):
        for j in xrange(1, matrix.shape[1]):
            c1 = str1[i - 1]
            c2 = str2[j - 1]

            scores = (('s', matrix[i - 1, j - 1] + charmatrix.get((c1, c2),
                                                                  substitutionscore)),
                      ('i', matrix[i, j - 1] + charmatrix.get((c1, c2),
                                                              insertscore)),
                      ('d', matrix[i - 1, j] + charmatrix.get((c1, c2),
                                                              deletescore)))
            if str1[i - 1] == str2[j - 1]:
                matrix[i, j] = matrix[i - 1, j - 1]
                steps[i, j] = 'm'
            else:
                bestoption = min(scores, key=lambda x: x[1])
                matrix[i, j] = bestoption[1]
                steps[i, j] = bestoption[0]

    return matrix, steps

# ----------------------------------------------------------------------
# Language algorithms --------------------------------------------------
# ----------------------------------------------------------------------

# Useful unicode rang
ascii_range = (u'Ascii', unichr(0), unichr(127))
greek_coptic_range = (u'Greek Coptic', u'\u0370', u'\u03FF')
extended_greek_range = (u'Extended Greek', u'\u1F00', u'\u1FFF')
combining_diacritical_mark_range = (
    u'Combining Diacritical', u'\u0300', u'\u036f')

# This is a list of all tone mark code points not in the combining
# diacritial block. They are from the "Greek and Coptic" and "Extended
# Greek" blocks.
greek_and_coptic_diacritics = [u'\u037A', u'\u0384', u'\u0385']
extended_greek_diacritics = [u'\u1fbd', u'\u1fbe', u'\u1fbf', u'\u1fc0',
                             u'\u1fc1', u'\u1fcd', u'\u1fce', u'\u1fcf',
                             u'\u1fdd', u'\u1fde', u'\u1fdf', u'\u1fed',
                             u'\u1fee', u'\u1fef', u'\u1ffd', u'\u1ffe']


def uniblock(start, stop):
    """
    Return a list containing all the characters in the unicode table
    starting with 'start' (inclusive) and ending with end (inclusive).
    """
    ints = range(ord(start), ord(stop) + 1)

    return map(unichr, ints)


def inblock(c, bounds):
    """
    Check that the character c is equal to or between the two bounding
    characters in the unicode table.
    """
    return ord(c) >= ord(bounds[0]) and ord(c) <= ord(bounds[1])


@unibarrier
def identify(string, unicode_blocks):
    """
    Determine percent-wise how many characters in the given string
    belong in each given unicode block. Ranges may be user defined, and
    not official unicode ranges. It is assumed that ranges do not
    overlap. unicode_blocks is an iterable of 3-tuples of the form
    (<name of block>, <first unichar in block>, <last unichar in the
    block>).
    """
    result = {b[0]: 0 for b in unicode_blocks}
    for c in string:
        for r in unicode_blocks:
            if inblock(c, (r[1], r[2])):
                result[r[0]] += 1

    return result


@unibarrier
def islang(unistr, unicode_blocks, threshold=1.0):
    """
    Determine if a given (unicode) string belongs to a certain langauge.
    This calls the identify function to determine what fraction of the
    string's characters are in the specified blocks. It returns true if
    the number of chars in those blocks is >= threshold.
    Threshold is a float between 0 and 1.
    """
    if threshold > 1.0 or threshold <= 0.0:
        raise Exception(u'Threshold must be > 0.0 and <= 1.0')

    res = identify(unistr, unicode_blocks)
    inlang = 0
    for block in unicode_blocks:
        inlang += res[block[0]]

    return inlang / len(unistr) >= threshold


@unibarrier
def isgreek(ustr):
    return islang(ustr, [greek_coptic_range, extended_greek_diacritics,
                         greek_and_coptic_diacritics])


def greek_chars():
    """
    Return a list containing all the characters from the Greek and
    Coptic, Extended Greek, and Combined Diacritical unicode blocks.
    """

    chars = uniblock(greek_coptic_range[1], greek_coptic_range[2])
    chars += uniblock(extended_greek_range[1], extended_greek_range[2])
    chars += uniblock(combining_diacritical_mark_range[1],
                      combining_diacritical_mark_range[2])
    return chars


@unibarrier
def greek_filter(string):
    """
    Remove all non-Greek characters from a string.
    """
    return filter(greek_chars().__contains__, string)


@unibarrier
def strip_diacritics(ustr):
    """
    Remove all Greek diacritics from the specified string. Expects the
    string to be in NFD.
    """
    diacritics = uniblock(combining_diacritical_mark_range[1],
                          combining_diacritical_mark_range[2])
    diacritics += greek_and_coptic_diacritics + extended_greek_diacritics
    return u''.join(c for c in ustr if c not in diacritics)


def list_to_uni(l, encoding=u'utf-8'):
    """
    Return a human readable string representation of a list of unicode
    strings using the specified encoding.
    """
    result = u'['
    for i in xrange(0, len(l)):
        result += l[i]
        if i != len(l) - 1:
            result += u', '
    result += u']'
    return result.encode(encoding)



import os
import urllib2
import urlparse
import shutil
from distutils.core import Command

manifest_url = "http://l.unchti.me/nidaba/MANIFEST"
download_prefix = "http://l.unchti.me/nidaba/"

class DownloadCommand(Command):
    command_name = 'download'
    description = ('Download misc. data files (dictionaries, sample inputs and'
                   'models for tests etc.)')
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print("Downloading manifest...")
        manifest = [x.strip() for x in
                    urllib2.urlopen(manifest_url).readlines()]
        print("Downloading: ")
        for f in manifest:
            print('\t* ' + f)
            try:
                os.makedirs(os.path.dirname(f))
            except OSError:
                pass
            r = urllib2.urlopen(urlparse.urljoin(download_prefix, f))
            with open(f, 'wb') as fp:
                shutil.copyfileobj(r, fp)



# -*- coding: utf-8 -*-
"""
nidaba.plugins.leptonica
~~~~~~~~~~~~~~~~~~~~~~~~

Plugin accessing `leptonica <http://leptonica.com>`_ functions.

This plugin requires a liblept shared object in the current library search
path. On Debian-based systems it can be installed using apt-get::

    # apt-get install libleptonica-dev

Leptonica's APIs are rather unstable and may differ significantly between
versions. If this plugin fails with weird error messages or workers are just
dying without discernable cause please submit a bug report including your
leptonica version.
"""

from __future__ import absolute_import

import ctypes

from nidaba import storage

from nidaba.celery import app
from nidaba.tasks.helper import NidabaTask
from nidaba.nidabaexceptions import (NidabaInvalidParameterException,
                                     NidabaLeptonicaException,
                                     NidabaPluginException)

leptlib = 'liblept.so'


def setup(*args, **kwargs):
    try:
        ctypes.cdll.LoadLibrary(leptlib)
    except Exception as e:
        raise NidabaPluginException(e.message)


@app.task(base=NidabaTask, name=u'nidaba.binarize.sauvola')
def sauvola(doc, method=u'sauvola', whsize=10, factor=0.35):
    """
    Binarizes an input document utilizing Sauvola thresholding as described in
    [0]. Expects 8bpp grayscale images as input.

    [0] Sauvola, Jaakko, and Matti Pietikäinen. "Adaptive document image
    binarization." Pattern recognition 33.2 (2000): 225-236.

    Args:
        doc (unicode): The input document tuple.
        method (unicode): The suffix string appended to all output files
        whsize (int): The window width and height that local statistics are
                      calculated on are twice the value of whsize. The minimal
                      value is 2.
        factor (float): The threshold reduction factor due to variance. 0 =<
                        factor < 1.

    Returns:
        (unicode, unicode): Storage tuple of the output file

    Raises:
        NidabaInvalidParameterException: Input parameters are outside the valid
                                         range.
    """
    input_path = storage.get_abs_path(*doc)
    output_path = storage.insert_suffix(input_path, method, unicode(whsize),
                                        unicode(factor))
    lept_sauvola(input_path, output_path, whsize, factor)
    return storage.get_storage_path(output_path)


def lept_sauvola(image_path, output_path, whsize=10, factor=0.35):
    """
    Binarizes an input document utilizing Sauvola thresholding as described in
    [0]. Expects 8bpp grayscale images as input.

    [0] Sauvola, Jaakko, and Matti Pietikäinen. "Adaptive document image
    binarization." Pattern recognition 33.2 (2000): 225-236.

    Args:
        image_path (unicode): Input image path
        output_path (unicode): Output image path
        whsize (int): The window width and height that local statistics are
                      calculated on are twice the value of whsize. The minimal
                      value is 2.
        factor (float): The threshold reduction factor due to variance. 0 =<
                        factor < 1.

    Raises:
        NidabaInvalidParameterException: Input parameters are outside the valid
                                         range.
    """

    if whsize < 2 or factor >= 1.0 or factor < 0:
        raise NidabaInvalidParameterException('Parameters (' + unicode(whsize)
                                              + ',' + unicode(factor) +
                                              ') outside of valid range')
    try:
        lept = ctypes.cdll.LoadLibrary(leptlib)
    except OSError as e:
        raise NidabaLeptonicaException('Loading leptonica failed: ' + e.message)
    pix = ctypes.c_void_p(lept.pixRead(image_path.encode('utf-8')))
    opix = ctypes.c_void_p()
    if lept.pixGetDepth(pix) != 8:
        lept.pixDestroy(ctypes.byref(pix))
        raise NidabaLeptonicaException('Input image is not grayscale')
    if lept.pixSauvolaBinarize(pix, whsize, ctypes.c_float(factor), 0, None,
                               None, None, ctypes.byref(opix)):
        lept.pixDestroy(ctypes.byref(pix))
        raise NidabaLeptonicaException('Binarization failed for unknown'
                                       'reason.')
    if lept.pixWriteImpliedFormat(output_path.encode('utf-8'), opix, 100, 0):
        lept.pixDestroy(ctypes.byref(pix))
        lept.pixDestroy(ctypes.byref(pix))
        raise NidabaLeptonicaException('Writing binarized PIX failed')
    lept.pixDestroy(ctypes.byref(opix))
    lept.pixDestroy(ctypes.byref(pix))


@app.task(base=NidabaTask, name=u'nidaba.img.dewarp')
def dewarp(doc, method=u'dewarp'):
    """
    Removes perspective distortion (as commonly exhibited by overhead scans)
    from an 1bpp input image.

    Args:
        doc (unicode, unicode): The input document tuple.
        method (unicode): The suffix string appended to all output files.

    Returns:
        (unicode, unicode): Storage tuple of the output file
    """
    input_path = storage.get_abs_path(*doc)
    output_path = storage.insert_suffix(input_path, method)
    lept_dewarp(input_path, output_path)
    return storage.get_storage_path(output_path)


def lept_dewarp(image_path, output_path):
    """
    Removes perspective distortion from an 1bpp input image.

    Args:
        image_path (unicode): Path to the input image
        output_path (unicode): Path to the output image

    Raises:
        NidabaLeptonicaException if one of leptonica's functions failed.
    """
    try:
        lept = ctypes.cdll.LoadLibrary(leptlib)
    except OSError as e:
        raise NidabaLeptonicaException('Loading leptonica failed: ' + e.message)
    pix = ctypes.c_void_p(lept.pixRead(image_path.encode('utf-8')))
    opix = ctypes.c_void_p()
    ret = lept.dewarpSinglePage(pix, 0, 1, 1, ctypes.byref(opix), None, 0)
    if ret == 1 or ret is None:
        lept.pixDestroy(ctypes.byref(pix))
        lept.pixDestroy(ctypes.byref(opix))
        raise NidabaLeptonicaException('Dewarping failed for unknown reason.')
    if lept.pixWriteImpliedFormat(output_path.encode('utf-8'), opix, 100, 0):
        lept.pixDestroy(ctypes.byref(pix))
        lept.pixDestroy(ctypes.byref(opix))
        raise NidabaLeptonicaException('Writing dewarped PIX failed')
    lept.pixDestroy(ctypes.byref(pix))
    lept.pixDestroy(ctypes.byref(opix))


@app.task(base=NidabaTask, name=u'nidaba.img.deskew')
def deskew(doc, method=u'deskew'):
    """
    Removes skew (rotational distortion) from an 1bpp input image.

    Args:
        doc (unicode, unicode): The input document tuple.
        method (unicode): The suffix string appended to all output files.

    Returns:
        (unicode, unicode): Storage tuple of the output file
    """
    input_path = storage.get_abs_path(*doc)
    output_path = storage.insert_suffix(input_path, method)
    lept_deskew(input_path, output_path)
    return storage.get_storage_path(output_path)


def lept_deskew(image_path, output_path):
    """
    Removes skew (rotational distortion from an 1bpp input image.

    Args:
        image_path (unicode): Input image
        output_path (unicode): Path to the output document

    Raises:
        NidabaLeptonicaException if one of leptonica's functions failed.
    """
    try:
        lept = ctypes.cdll.LoadLibrary(leptlib)
    except OSError as e:
        raise NidabaLeptonicaException('Loading leptonica failed: ' + e.message)
    pix = ctypes.c_void_p(lept.pixRead(image_path.encode('utf-8')))
    opix = ctypes.c_void_p(lept.pixFindSkewAndDeskew(pix, 4, None, None))
    if opix is None:
        lept.pixDestroy(ctypes.byref(pix))
        raise NidabaLeptonicaException('Deskewing failed for unknown reason.')
    if lept.pixWriteImpliedFormat(output_path.encode('utf-8'), opix, 100, 0):
        lept.pixDestroy(ctypes.byref(pix))
        lept.pixDestroy(ctypes.byref(opix))
        raise NidabaLeptonicaException('Writing deskewed PIX failed')
    lept.pixDestroy(ctypes.byref(pix))
    lept.pixDestroy(ctypes.byref(opix))



# -*- coding: utf-8 -*-
"""
nidaba.plugins.tesseract
~~~~~~~~~~~~~~~~~~~~~~~~

Plugin implementing an interface to tesseract

This plugin exposes tesseract's functionality as a task. It implements two ways
of calling tesseract, a direct method calling the tesseract executable and one
utilizing the C-API available from tesseract 3.02 and upwards.

The C-API requires a libtesseract shared object in the current library path and
training data in the configured tessdata directory:

.. code-block:: console

    # apt-get install libtesseract3 tesseract-ocr-$lang

Using the direct call method requires the tesseract binary installable by
executing:

.. code-block:: console

    # apt-get install tesseract-ocr

.. note::
    It is strongly encouraged to use the C-API whenever possible. It is
    supposedly stable while hOCR output file names change between tesseract
    versions.
    
Configuration
~~~~~~~~~~~~~

implementation (default='capi')
    Selector for the call method. May either be `capi`, `direct` (tesseract
    hOCR output with .hocr extension), or `legacy` (tesseract hOCR output with
    .html extension).

tessdata (default='/usr/share/tesseract-ocr/')
    Path to load tesseract training data and configuration from. Has to be one
    directory level upwards from the actual tessdata directory
"""

from __future__ import absolute_import

import subprocess
import ctypes

from PIL import Image
from distutils import spawn

from nidaba import storage
from nidaba.celery import app
from nidaba.tasks.helper import NidabaTask
from nidaba.nidabaexceptions import NidabaTesseractException
from nidaba.nidabaexceptions import NidabaPluginException


implementation = u'capi'
tessdata = u'/usr/share/tesseract-ocr/'


def setup(*args, **kwargs):
    if kwargs.get(u'implementation'):
        global implementation
        implementation = kwargs.get(u'implementation')
    if kwargs.get(u'tessdata'):
        global tessdata
        if isinstance(kwargs.get(u'tessdata'), list):
            tessdata = storage.get_abs_path(*kwargs.get(u'tessdata'))
        else:
            tessdata = kwargs.get(u'tessdata')
    if implementation == 'direct' and not spawn.find_executable('tesseract'):
        raise NidabaPluginException('No tesseract executable found')
    if implementation == 'capi' :
        try:
            ctypes.cdll.LoadLibrary('libtesseract.so.3')
        except:
            raise NidabaPluginException('Loading libtesseract failed.')


@app.task(base=NidabaTask, name=u'nidaba.ocr.tesseract')
def ocr_tesseract(doc, method=u'ocr_tesseract', languages=None):
    """
    Runs tesseract on an input document.

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string appended to all output files
                          languages (list of unicode): A list of languages for
                          the tesseract language model

    Returns:
        (unicode, unicode): Storage tuple for the output file
    """
    image_path = storage.get_abs_path(*doc)
    if isinstance(languages, basestring):
        languages = [languages]
    output_path = storage.insert_suffix(image_path, method, *languages)

    if implementation == 'legacy':
        result_path = output_path + '.html'
        ocr_direct(image_path, output_path, languages)
    elif implementation == 'direct':
        result_path = output_path + '.hocr'
        ocr_direct(image_path, output_path, languages)
    elif implementation == 'capi':
        result_path = output_path + '.hocr'
        ocr_capi(image_path, result_path, languages)
    else:
        raise NidabaTesseractException('Invalid implementation selected',
                                       implementation)
    return storage.get_storage_path(result_path)


def ocr_capi(image_path, output_path, languages):
    """
    OCRs an image using the C API provided by tesseract versions 3.02 and
    higher. Images are read using pillow allowing a wider range of formats than
    leptonica and results are written to a fixed output document.

    Args:
        image_path (unicode): Path to the input image
        output_path (unicode): Path to the hOCR output
        languages (list): List of valid tesseract language identifiers
    """

    img = Image.open(image_path)
    w, h = img.size

    assert img.mode == 'L' or img.mode == '1'
    # There is a regression in TessBaseAPISetImage somewhere between 3.02 and
    # 3.03. Unfortunately the tesseract maintainers close all bug reports
    # concerning their API, so we just convert to grayscale here.
    img = img.convert('L')
    try:
        tesseract = ctypes.cdll.LoadLibrary('libtesseract.so.3')
    except OSError as e:
        raise NidabaTesseractException('Loading libtesseract failed: ' +
                                       e.message)

    # ensure we've loaded a tesseract object newer than 3.02
    tesseract.TessVersion.restype = ctypes.c_char_p
    ver = tesseract.TessVersion()
    if int(ver.split('.')[0]) < 3 or int(ver.split('.')[1]) < 2:
        tesseract.TessBaseAPIDelete(api)
        raise NidabaTesseractException('libtesseract version is too old. Set'
                                       'implementation to direct.')
    api = tesseract.TessBaseAPICreate()
    rc = tesseract.TessBaseAPIInit3(api, str(tessdata), str('+'.join(languages)))
    if (rc):
        tesseract.TessBaseAPIDelete(api)
        raise NidabaTesseractException('Tesseract initialization failed.')
    tesseract.TessBaseAPISetImage(api, ctypes.c_char_p(str(img.tobytes())), w, h, 1, w)
    with open(output_path, 'wb') as fp:
        tp = tesseract.TessBaseAPIGetHOCRText(api)
        fp.write(ctypes.string_at(tp))
        tesseract.TessDeleteText(tp)
    tesseract.TessBaseAPIDelete(api)

def ocr_direct(image_path, output_path, languages):
    """
    OCRs an image by calling the tesseract executable directly. Images are read
    using the linked leptonica library and the given output_path WILL be
    modified by tesseract.

    Args:
        image_path (unicode): Path to the input image
        output_path (unicode): Path to the hOCR output
        languages (list): List of valid tesseract language identifiers
    """

    p = subprocess.Popen(['tesseract', image_path, output_path, '-l',
                          '+'.join(languages), '--tessdata-dir', tessdata,
                          'hocr'], stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate()
    if p.returncode:
        raise NidabaTesseractException(err)



# -*- coding: utf-8 -*-
"""
nidaba.tasks.stats
~~~~~~~~~~~~~~~~~~

Various tasks calculating metrics on documents.

"""

from __future__ import absolute_import, unicode_literals

import os

from nidaba.tasks.helper import NidabaTask
from nidaba.algorithms import string
from nidaba.celery import app
from nidaba import storage


@app.task(base=NidabaTask, name=u'nidaba.stats.text_error_rate')
def text_error_rate(doc, method=u'error_rate', ground_truth=None, divert=True):
    """
    Estimates the error rate (edit distance) between an input document and a
    given ground truth.

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string appended to the output file.
        ground_truth (unicode): Ground truth location tuple
        divert (bool): Switch selecting output diversion. If enabled the output
                       will be added to the tracking arguments and the input
                       document will be returned as the result of the task. Use
                       this to insert a statistical measure into a chain
                       without affecting the results.

    Returns:
        (unicode, unicode): Storage tuple of the output document
    """
    input_path = storage.get_abs_path(*doc[0])

    output_path = storage.insert_suffix(input_path, method,
                                        os.path.basename(input_path))
    edist = string.edit_distance(storage.retrieve_text(*doc),
                                 storage.retrieve_text(*ground_truth))
    print(edist)
    if not divert:
        return output_path
    else:
        return {'error_rate': (doc, edist), 'doc': doc}



# -*- coding: utf-8 -*-
"""
nidaba.lex
~~~~~~~~~~

This module contains functions for dealing with words and dictionaries, such as
extracting words from texts, normalizing encodings, building symmetric deletion
dictionaries, etc.
"""

from __future__ import absolute_import

import os
import codecs
import glob
import nidaba.algorithms.string as alg
from collections import Counter


@alg.unibarrier
def cleanlines(path, encoding=u'utf-8', normalization=u'NFD'):
    """
    Read in lines from a file and return them as a sanitized list.
    Non-unique linse will be repeated.

    Args:
        path (unicode): Absolute path of the file to be read
        encoding (unicode): Encoding to use for decoding the file
        normalization (unicode): Normalization format to use

    Returns:
        list: List of lines containing the sanitized output, i.e. normalized
              unicode objects.
    """
    words = []
    with codecs.open(path, u'r', encoding=encoding) as lines:
        for line in lines:
            words.append(alg.sanitize(line, normalization=normalization))
    return words


@alg.unibarrier
def cleanwords(path, encoding=u'utf-8', normalization=u'NFD'):
    """
    Read in every word from a files as separated by lines and spaces.
    Non-unique words will be repeated as they are read in. Detects only words
    divided by a standard space.

    Args:
        path (unicode): Absolute path of the file to be read
        encoding (unicode): Encoding to use for decoding the file
        normalization (unicode): Normalization format to use

    Returns:
        list: List of words containing the sanitized output, i.e. normalized
              unicode objects.
    """
    words = []
    with codecs.open(path, u'r', encoding=encoding) as lines:
        for line in lines:
            for seg in line.split(u' '):
                clean = alg.sanitize(seg, normalization=normalization)
                if clean != u'':
                    words.append(clean)
    return words


@alg.unibarrier
def uniquewords_with_freq(path, encoding=u'utf-8', normalization=u'NFD'):
    """
    Read in every word from a file as separated by lines and spaces.
    Return a counter (behaves like a dictionary) of unique words
    along with the number of times they occurred.

    Args:
        path (unicode): Absolute path of the file to be read
        encoding (unicode): Encoding to use for decoding the file
        normalization (unicode): Normalization format to use

    Returns:
        Counter: Contains the frequency of each token
    """
    words = cleanwords(path, encoding=encoding, normalization=normalization)
    freq = Counter(words)
    return freq


@alg.unibarrier
def cleanuniquewords(path, encoding=u'utf-8', normalization=u'NFD'):
    """
    Read in lines from a file as separated by lines and spaces,
    convert them to the specified normalization, and return a set
    of all unique words.

    Args:
        path (unicode): Absolute path of the file to be read
        encoding (unicode): Encoding to use for decoding the file
        normalization (unicode): Normalization format to use

    Returns:
        set: Set of unique tokens
    """
    return set(cleanwords(path, encoding=encoding,
                          normalization=normalization))


@alg.unibarrier
def words_from_files(dirpath, encoding=u'utf-8', normalization=u'NFD'):
    """
    Create a dictionary from a directory of text files.  All files in the given
    directory will be parsed.

    Args:
        dirpath (unicode): Absolute path of the directory to enter
        encoding (unicode): Encoding to use for decoding the files
        normalization (unicode): Normalization format to use

    Returns:
        list: List of words of all files in the directory
    """
    words = []
    for filename in filter(os.path.isfile, glob.glob(dirpath + '/*')):
        words += cleanwords(filename, encoding=encoding,
                            normalization=normalization)
    return words


@alg.unibarrier
def unique_words_from_files(dirpath, encoding=u'utf-8', normalization=u'NFD'):
    """
    Create a set of unique words from a directory of text files.  All file in
    the given directory will be parsed.

    Args:
        dirpath (unicode): Absolute path of the directory to enter
        encoding (unicode): Encoding to use for decoding the files
        normalization (unicode): Normalization format to use

    Returns:
        set: Set of words of all files in the directory
    """
    return set(words_from_files(dirpath, encoding=encoding,
                                normalization=normalization))


@alg.unibarrier
def make_dict(outpath, iterable, encoding=u'utf-8'):
    """
    Create a file at outpath and write evrey object in iterable to its
    own line. The file is opened in append mode.

    Args:
        outpath (unicode): File path to write to
        iterable (iterable): An iterable used as a data source
        normalization (unicode): Normalization format to use
    """
    with codecs.open(outpath, u'w+', encoding=encoding) as f:
        for s in iterable:
            f.write(s + u'\n')


@alg.unibarrier
def make_deldict(outpath, words, depth):
    """
    Creates a symmetric deletion dictionary from the specified word list.

    WARNING! This is a naive approach, which requires all the variants to be
    stored in memory. For large dictionaries at higher depth, this can easily
    use all available memory on most machines.

    Args:
        outpath (unicode): File path to write to
        words (iterable): An iterable returning a single word per iteration
        depth (int): Maximum edit distance to calculate
    """
    variant_dict = {}
    for word in words:
        for var in alg.strings_by_deletion(word, depth):
            if var not in variant_dict:
                variant_dict[var] = []
            variant_dict[var].append(word)
    ordered = sorted(variant_dict.keys())

    with codecs.open(outpath, u'w+', encoding='utf-8') as outfile:
        for key in ordered:
            originals = u' '.join(variant_dict[key])
            outfile.write(u'%s\t%s' % (key, originals) + u'\n')



# -*- coding: utf-8 -*-
import unittest
import os
import shutil
import tempfile
import os

from lxml import etree
from distutils import spawn
from nose.plugins.skip import SkipTest
from nidaba.plugins import ocropus

thisfile = os.path.abspath(os.path.dirname(__file__))
resources = os.path.abspath(os.path.join(thisfile, 'resources/ocropus'))


class OcropusTests(unittest.TestCase):

    """
    Tests for python ocropus bindings.
    """

    def setUp(self):
        if None in [spawn.find_executable('ocropus-rpred'),
                    spawn.find_executable('ocropus-gpageseg'),
                    spawn.find_executable('ocropus-hocr')]:
            raise SkipTest

        self.otempdir = unicode(tempfile.mkdtemp())
        # copytree fails if the target directory already exists. Unfortunately
        # not creating the temporary directory using mkstemp() is
        # race-conditiony.
        self.tempdir = os.path.join(self.otempdir, u'stupid')
        shutil.copytree(resources, self.tempdir)

    def tearDown(self):
        shutil.rmtree(self.otempdir)

    def test_file_path_correct(self):
        """
        Test that output is placed in the correct directory.
        """
        pngpath = os.path.join(self.tempdir, u'image_png.png')
        wd = tempfile.mkdtemp()
        swd = tempfile.mkdtemp()

        outpath = os.path.join(swd, u'outpath_png.hocr')
        os.chdir(wd)

        modelpath = os.path.join(self.tempdir, u'en-default.pyrnn.gz')
        ocr = ocropus.ocr(pngpath, outpath, modelpath)
        self.assertTrue(os.path.isfile(ocr),
                        msg='Ocropus did not outpath a file!')
        self.assertEqual(os.path.dirname(ocr), swd,
                         msg='Output not placed in correct directory')
        try:
            etree.parse(ocr)
        except etree.XMLSyntaxError:
            self.fail(msg='The outpath was not valid html/xml!')
        finally:
            shutil.rmtree(swd)
            shutil.rmtree(wd)

    def test_file_outpath_png(self):
        """
        Test that ocropus creates hocr output for pngs.
        """
        pngpath = os.path.join(self.tempdir, u'image_png.png')
        outpath = os.path.join(self.tempdir, u'outpath_png.hocr')
        modelpath = os.path.join(self.tempdir, u'en-default.pyrnn.gz')
        ocr = ocropus.ocr(pngpath, outpath, modelpath)
        self.assertTrue(os.path.isfile(ocr),
                        msg='Ocropus did not outpath a file!')
        try:
            etree.parse(ocr)
        except etree.XMLSyntaxError:
            self.fail(msg='The outpath was not valid html/xml!')

    def test_file_outpath_tiff(self):
        """
        Test that ocropus creates hocr output for tiffs.
        """
        tiffpath = os.path.join(self.tempdir, u'image_tiff.tiff')
        outpath = os.path.join(self.tempdir, u'outpath_tiff.hocr')
        modelpath = os.path.join(self.tempdir, u'en-default.pyrnn.gz')
        ocr = ocropus.ocr(tiffpath, outpath, modelpath)
        self.assertTrue(os.path.isfile(ocr),
                        msg='Ocropus did not output a file!')
        try:
            etree.parse(ocr)
        except etree.XMLSyntaxError:
            self.fail(msg='The outpath was not valid html/xml!')

    def test_file_outpath_jpg(self):
        """
        Test that ocropus creates hocr output for jpgs.
        """
        jpgpath = os.path.join(self.tempdir, u'image_jpg.jpg')
        outpath = os.path.join(self.tempdir, u'outpath_jpg.hocr')
        modelpath = os.path.join(self.tempdir, u'en-default.pyrnn.gz')
        ocr = ocropus.ocr(jpgpath, outpath, modelpath)
        self.assertTrue(os.path.isfile(ocr),
                        msg='Ocropus did not output a file!')
        try:
            etree.parse(ocr)
        except etree.XMLSyntaxError:
            self.fail(msg='The outpath was not valid html/xml!')


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import codecs
from nidaba.algorithms import strings_by_deletion, sanitize


def file_len(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1

if __name__ == '__main__':
    """
    Generate a dictionary for symmetric deletion spell checking.
    Lines in the dictionary are of the form:
    variant : [list of originals]
    """
    if len(sys.argv) != 4:
        print u'mkdict [word list] [dictionary file] [edit distance]'
        exit()
    elif sys.argv[3] <= 0:
        print 'Error depth must be positive'
    input_file_path = sys.argv[1]
    output_file_path = sys.argv[2]
    edit_distance = int(sys.argv[3])

    fl = file_len(input_file_path)

    with open(input_file_path) as infile:
        i = 0
        dic = {}
        for word in infile:
            word = sanitize(word)
            variants = strings_by_deletion(word, edit_distance)
            i += 1
            sys.stdout.write("Words generated:%d/%i   \r" % (i, fl))
            sys.stdout.flush()

            for v in variants:
                if v not in dic:
                    dic[v] = []
                dic[v].append(word)

    print 'sorting...'
    ordered = sorted(dic.keys())

    print '\n'
    print 'writing file'
    with codecs.open(output_file_path, u'w+', encoding='utf-8') as outfile:
        i = 0
        for k in ordered:
            sys.stdout.write("Entries written:%d/%i   \r" % (i, len(dic) - 1))
            if i != fl - 1:
                sys.stdout.flush()

            originals = u' '.join(dic[k])
            outfile.write(u'%s : %s' % (k, originals) + u'\n')
            i += 1
    print '\n'



# -*- coding: utf-8 -*-
"""
nidaba.tasks.helper
~~~~~~~~~~~~~~~~~~~

A helper class that all nidaba tasks should inherit from to ensure accurate
logging of errors.
"""

from __future__ import absolute_import

from celery import Task
from inspect import getargspec
from nidaba.celery import app

import json


class NidabaTask(Task):
    """
    An abstract class propagating unused function arguments through the
    execution chain. This means that no task should accept arbitrary
    (\*\*kwargs) arguments as they won't be forwarded to the actual function
    and will be retained through the whole chain.
    """
    abstract = True
    acks_late = True

    def __call__(self, *args, **kwargs):
        # if args is a dictionary we merge it into kwargs
        if len(args) == 1 and isinstance(args[0], dict):

            kwargs.update(args[0])
            args = ()
            # after a step the output of all tasks is merged into a single
            # argument (indicated by the doc argument being a list of dicts)
            # which we have to unravel first before it can be used be
            # subsequent tasks.
            if isinstance(kwargs['doc'][0], dict):
                docs = []
                root = set()
                for o in kwargs['doc']:
                    docs.append(tuple(o['doc']))
                    root.add(tuple(o['root']))
                kwargs['doc'] = docs
                assert len(root) == 1, 'Nonmatching root documents'
                kwargs['root'] = root.pop()

        # and then filter all tracking objects (root document, job id, ...) out
        # again
        fspec = getargspec(self.run)
        nkwargs = {}
        tracking_kwargs = {}
        while kwargs:
            k, v = kwargs.popitem()
            if k in fspec.args:
                nkwargs[k] = v
            else:
                tracking_kwargs[k] = v
        try:
            ret = super(NidabaTask, self).__call__(*args, **nkwargs)
        except Exception as e:
            # write error to backend and reraise exception
            batch_struct = json.loads(app.backend.get(tracking_kwargs['id']))
            batch_struct['errors'].append((nkwargs, tracking_kwargs,
                                           e.message))
            app.backend.set(tracking_kwargs['id'], json.dumps(batch_struct))
            raise
        if isinstance(ret, dict):
            for k,v in ret.iteritems():
                # if a previous task has already created the key, for example a
                # measure being computed before and after a processing step, we
                # append the value.
                if k in tracking_kwargs:
                    tracking_kwargs[k].append(v)
                else:
                    tracking_kwargs[k] = v
        else:
            tracking_kwargs['doc'] = ret
        return tracking_kwargs



# -*- coding: utf-8 -*-
"""
nidaba.tasks.postprocessing
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Various postprocessing tasks that operate upon recognized texts.

"""

from __future__ import absolute_import, unicode_literals

from nidaba.tasks.helper import NidabaTask
from nidaba.celery import app
from nidaba import storage
from nidaba import merge_hocr


@app.task(base=NidabaTask, name=u'nidaba.postprocessing.blend_hocr')
def blend_hocr(doc, method=u'blend_hocr', language=u''):
    """
    Blends multiple hOCR files using the algorithm from Bruce Robertsons
    rigaudon. It requires a working spell checking for the input document's
    language; otherwise all matched bboxes will be bunched together without any
    scoring.

    Args:
        doc [(id, path), ...]: A list of storage module tupels that will be
        merged into a single output document.
        language (unicode): Language used for spell-checking based scoring. If
                            not defined no scoring will be used.
        method (unicode): The suffix string appended to the output file.

    Returns:
        (unicode, unicode): Storage tuple of the output document
    """
    # create the output document path from the first input document
    input_path = storage.get_abs_path(*doc[0])
    output_path = storage.insert_suffix(input_path, method)
    return merge_hocr.merge(doc, language,
                            storage.get_storage_path(output_path))



# -*- coding: utf-8 -*-
import unittest
import os
import tempfile
import shutil
from nidaba import lex


class DictTests(unittest.TestCase):

    """
    General tests for the dict.py module.
    """

    def setUp(self):
        # Make a temp sample dictionary
        self.temp = tempfile.NamedTemporaryFile()
        self.temp.write(u'word1\n'.encode(u'utf-8'))
        self.temp.write(u'word2\n'.encode(u'utf-8'))
        self.temp.write(u'αχιλλεύς\n'.encode(u'utf-8'))  # Achilles, in NFD
        self.temp.write(u'αχιλλεύς\n'.encode(u'utf-8'))  # Achilles, in NFC
        self.temp.seek(0, 0)
        self.path = os.path.abspath(self.temp.name).decode(u'utf-8')

        # Make an empty temp directory for general use
        self.tempdir = tempfile.mkdtemp().decode('utf-8')

    def tearDown(self):
        self.temp.close()
        shutil.rmtree(self.tempdir)

    def test_cleanlines(self):
        """
        Test the cleanline function.
        """
        words = lex.cleanlines(self.path)
        self.assertEqual(len(words), 4)
        self.assertEqual(words[0], u'word1')
        self.assertEqual(words[1], u'word2')
        self.assertEqual(words[2], u'αχιλλεύς')
        self.assertEqual(words[3], u'αχιλλεύς')

    def test_cleanwords(self):
        """
        Test the cleanwords function.
        """
        self.temp.write(
            u'adding a line with multiple words\n'.encode(u'utf-8'))
        self.temp.write(u'another with Greek αχιλλεύς\n'.encode(u'utf-8'))
        self.temp.write(u'and some NFC αχιλλεύς\n'.encode(u'utf-8'))
        self.temp.seek(0, 0)
        expected = [u'adding', u'a', u'line', u'with', u'multiple', u'words',
                    u'another', u'with', u'Greek', u'αχιλλεύς', u'and',
                    u'some', u'NFC', u'αχιλλεύς']
        words = lex.cleanwords(self.path)
        self.assertEqual(words, expected)

    def test_cleanuniquewords(self):
        """
        Test the cleanuniquewords function.
        """
        self.temp1 = tempfile.NamedTemporaryFile()
        self.temp1.write(
            u'adding a line with multiple words\n'.encode(u'utf-8'))
        self.temp1.write(u'another with Greek αχιλλεύς\n'.encode(u'utf-8'))
        self.temp1.write(u'and some NFC αχιλλεύς\n'.encode(u'utf-8'))
        self.temp1.seek(0, 0)
        expected = set([u'adding', u'a', u'line', u'with', u'multiple',
                        u'words', u'another', u'with', u'Greek', u'αχιλλεύς',
                        u'and', u'some', u'NFC', u'αχιλλεύς'])
        self.assertEqual(
            lex.cleanuniquewords(self.temp1.name.decode(u'utf-8')), expected)

    def test_words_from_files(self):
        """
        Test the words_from_files function.
        """
        self.temp1 = tempfile.NamedTemporaryFile(dir=self.tempdir, delete=True)
        self.temp2 = tempfile.NamedTemporaryFile(dir=self.tempdir, delete=True)
        self.temp3 = tempfile.NamedTemporaryFile(dir=self.tempdir, delete=True)
        self.dirtofilter = tempfile.mkdtemp(
            dir=self.tempdir)  # To be filtered out
        self.temp1.write(u'a')
        self.temp2.write(u'b')
        self.temp3.write(u'c')
        self.temp1.seek(0, 0)
        self.temp2.seek(0, 0)
        self.temp3.seek(0, 0)
        words = lex.words_from_files(self.tempdir)
        self.assertEqual(3, len(words))
        self.assertTrue(u'a' in words)
        self.assertTrue(u'b' in words)
        self.assertTrue(u'c' in words)
        self.temp1.close()
        self.temp2.close()
        self.temp3.close()

    def test_unique_words_from_files(self):
        """
        Test the words_from_files_function.
        """
        self.temp1 = tempfile.NamedTemporaryFile(dir=self.tempdir, delete=True)
        self.temp2 = tempfile.NamedTemporaryFile(dir=self.tempdir, delete=True)
        self.temp3 = tempfile.NamedTemporaryFile(dir=self.tempdir, delete=True)
        self.temp4 = tempfile.NamedTemporaryFile(dir=self.tempdir, delete=True)
        self.dirtofilter = tempfile.mkdtemp(
            dir=self.tempdir)  # To be filtered out
        self.temp1.write(u'a')
        self.temp2.write(u'b b')
        self.temp3.write(u'c a')
        self.temp4.write(u'a b c \nd')
        self.temp1.seek(0, 0)
        self.temp2.seek(0, 0)
        self.temp3.seek(0, 0)
        self.temp4.seek(0, 0)
        words = lex.unique_words_from_files(self.tempdir)
        self.assertEqual(4, len(words))
        self.assertTrue(u'a' in words)
        self.assertTrue(u'b' in words)
        self.assertTrue(u'c' in words)
        self.assertTrue(u'd' in words)
        self.temp1.close()
        self.temp2.close()
        self.temp3.close()
        self.temp4.close()

    def test_make_dict(self):
        """
        Test the make_dict function.
        """
        words = [u'a', u'b', u'c', u'd']
        mytemp = tempfile.mkdtemp().decode('utf-8')
        outpath = os.path.join(mytemp, u'testdict')
        lex.make_dict(outpath, words)
        retrived = lex.cleanuniquewords(outpath)
        self.assertEqual(len(retrived), 4)
        self.assertTrue(u'a' in retrived)
        self.assertTrue(u'b' in retrived)
        self.assertTrue(u'c' in retrived)
        self.assertTrue(u'd' in retrived)
        shutil.rmtree(mytemp)

    def test_make_deldict_1(self):
        """
        Test the make_deldict function at depth 1.
        """
        outfile = tempfile.NamedTemporaryFile(dir=self.tempdir, delete=True)
        words = [u'aaa', u'bbb', u'ccc', u'ddd']
        lex.make_deldict(outfile.name, words, 1)
        lines = lex.cleanlines(outfile.name)
        self.assertEqual(u'aa\taaa', lines[0])
        self.assertEqual(u'bb\tbbb', lines[1])
        self.assertEqual(u'cc\tccc', lines[2])
        self.assertEqual(u'dd\tddd', lines[3])
        outfile.close()

    def test_unique_words_with_frequency(self):
        """
        Test the uniquewords_with_freq function
        """
        temp = tempfile.NamedTemporaryFile()
        temp.write('a a a a b b c c c b b b b d \n foobar foobar')
        temp.seek(0, 0)
        f = lex.uniquewords_with_freq(temp.name.decode('utf-8'))
        self.assertEqual(5, len(f.keys()))
        self.assertEqual(4, f['a'])
        self.assertEqual(6, f['b'])
        self.assertEqual(3, f['c'])
        self.assertEqual(1, f['d'])
        self.assertEqual(2, f['foobar'])

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import unittest

from nidaba import nidaba


class BatchTest(unittest.TestCase):

    """Tests the functionality of the Batch class."""

    def test_no_tick(self):
        pass

    def test_no_step(self):
        pass

    def test_no_task(self):
        pass

    def test_no_doc(self):
        pass

    def test_multiple_docs(self):
        pass

    def test_multiple_tasks(self):
        pass

    def test_multiple_ticks(self):
        pass

    def test_multiple_steps(self):
        pass

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
nidaba.tasks.img
~~~~~~~~~~~~~~~~

Some general image processing tasks that are outside the scope of more specific
packages (e.g. binarization).

"""

from __future__ import absolute_import, unicode_literals

from nidaba import storage
from nidaba import image
from nidaba.celery import app
from nidaba.tasks.helper import NidabaTask


@app.task(base=NidabaTask, name=u'nidaba.img.rgb_to_gray')
def rgb_to_gray(doc, method=u'rgb_to_gray'):
    """
    Converts an arbitrary bit depth image to grayscale and writes it back
    appending a suffix.

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string appended to all output files.

    Returns:
        (unicode, unicode): Storage tuple of the output file
    """
    input_path = storage.get_abs_path(*doc)
    output_path = storage.insert_suffix(input_path, method)
    return storage.get_storage_path(image.rgb_to_gray(input_path, output_path))



# -*- coding: utf-8 -*-
"""
nidaba.plugins.ocropus
~~~~~~~~~~~~~~~~~~~~~~

Plugin implementing an interface to the `ocropus
<http://github.com/tmbdev/ocropy>`_ OCR engine.

It requires working ocropus-* tools in your execution path. Please have a look
at the website for installation instructions. 

.. important::
    If you are not requiring specific functionality of ocropus please consider
    using the :mod:`kraken <nidaba.plugins.kraken>` plugin. Kraken does not
    require working around oddities in input argument acceptance and is
    generally more well-behaved than ocropus.

.. note::
    ocropus uses a peculiar implementation of splitext removing all characters
    after the first dot to determine the output path of OCR results. We
    mitigate this by appending a unique UUID for each input file before the
    first dot. Unfortunately, this process causes the loss of all processing
    info carried in the file name.
"""

from __future__ import absolute_import

import subprocess
import glob
import os
import re
import shutil
import uuid

from distutils import spawn
from nidaba import storage
from nidaba.config import nidaba_cfg
from nidaba.celery import app
from nidaba.tasks.helper import NidabaTask
from nidaba.nidabaexceptions import NidabaOcropusException
from nidaba.nidabaexceptions import NidabaPluginException

def setup(*args, **kwargs):
    if None in [spawn.find_executable('ocropus-rpred'),
                spawn.find_executable('ocropus-gpageseg'),
                spawn.find_executable('ocropus-hocr')]:
        raise NidabaPluginException('Prerequisites for ocropus module not installed.')

@app.task(base=NidabaTask, name=u'nidaba.ocr.ocropus')
def ocr_ocropus(doc, method=u'ocr_ocropus', model=None):
    """
    Runs ocropus on an input document.

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string appended to all output files
        model (unicode): Identifier for the font model to use

    Returns:
        (unicode, unicode): Storage tuple for the output file

    """
    input_path = storage.get_abs_path(*doc)
    output_path = os.path.splitext(storage.insert_suffix(input_path, method,
                                                         model))[0] + '.html'
    model = storage.get_abs_path(*(nidaba_cfg['ocropus_models'][model]))
    return storage.get_storage_path(ocr(input_path, output_path, model))


def _allsplitext(path):
    """
    Split all the pathname extensions, so that "a/b.c.d" -> "a/b", ".c.d"

    Args:
        path (unicode): A unicode object containing a file path.

    Returns:
        (tuple): Result containing firstly the path without extensions and
                 secondly the extracted extensions.
    """
    match = re.search(ur'((.*/)*[^.]*)([^/]*)', path)
    if not match:
        return (path, u'')
    else:
        return (match.group(1), match.group(3))


def ocr(imagepath, outputfilepath, modelpath):
    """
    Scan a single image with ocropus.

    Reads a single image file from ```imagepath``` and writes the recognized
    text as in hOCR format into outputfilepath. Ocropus's superfluous
    intermediate output remains in the directory imagepath is located in.

    Args:
        imagepath (unicode): Path of the input file
        outputfilepath (unicode): Path of the output file
        modelpath (unicode): Path of the recognition model. Must be a pyrnn.gz
                             pickle dump interoperable with ocropus-rpred.

    Returns:
        (unicode): A string of the output file that is actually written. As
                   Ocropus rewrites output file paths without notice it may be
                   different from the ```outputfilepath``` argument.

    Raises:
        NidabaOcropusException: Ocropus somehow failed. The error output is
                                contained in the message but as it is de facto
                                unusable as a library it's impossible to deduct
                                the nature of the problem.
    """

    fglob = _allsplitext(imagepath)[0]
    # ocropus removes everything after the first . anyway so we do it
    # preemptively here and add a stable unique identifier to ensure uniqueness
    # of output.
    outputfilepath = _allsplitext(outputfilepath)[0]
    outputfilepath += '_' + str(uuid.uuid5(uuid.NAMESPACE_URL,
                                           outputfilepath.encode('utf-8')))
    working_dir = os.path.dirname(outputfilepath)

    # ocropus is stupid and needs the input file to end in .bin.png
    shutil.copyfile(imagepath, fglob + '.bin.png')
    imagepath = fglob + '.bin.png'
    env = os.environ.copy()
    env['PYTHONIOENCODING'] = 'utf-8'
    # page layout analysis
    p = subprocess.Popen(['ocropus-gpageseg', '-n', imagepath.encode('utf-8')],
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         env=env, cwd=working_dir)
    out, err = p.communicate()
    print(out)
    if p.returncode:
        raise NidabaOcropusException(err)

    # text line recognition
    p = subprocess.Popen(['ocropus-rpred', '-q', '-m',
                          modelpath.encode('utf-8')] + glob.glob(fglob +
                                                                 '/*.bin.png'),
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         env=env, cwd=working_dir)
    out, err = p.communicate()
    if p.returncode:
        raise NidabaOcropusException(err)

    # recognition hOCR assembly
    # page segmentation is always converted to PNG
    p = subprocess.Popen(['ocropus-hocr', imagepath.encode('utf-8'), '-o',
                          outputfilepath.encode('utf-8')],
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         env=env, cwd=working_dir)
    out, err = p.communicate()
    if p.returncode:
        raise NidabaOcropusException(err)
    return outputfilepath



# -*- coding: utf-8 -*-
import unittest
import os
import shutil
import tempfile
import os

from lxml import etree
from nose.plugins.skip import SkipTest

thisfile = os.path.abspath(os.path.dirname(__file__))
resources = os.path.abspath(os.path.join(thisfile, 'resources/ocropus'))


class KrakenTests(unittest.TestCase):

    """
    Tests for python ocropus bindings.
    """

    def setUp(self):
        try:
            global kraken
            from nidaba.plugins import kraken
            kraken.setup()
            self.tempdir = unicode(tempfile.mkdtemp())
        except:
            raise SkipTest

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_file_outpath_png(self):
        """
        Test that kraken creates hocr output for pngs.
        """
        pngpath = os.path.join(resources, u'image_png.png')
        modelpath = os.path.join(resources, u'en-default.pyrnn.gz')
        ocr = kraken.ocr(pngpath, modelpath)
        try:
            parser = etree.HTMLParser()
            etree.fromstring(ocr, parser)
        except etree.XMLSyntaxError as e:
            print(e.message)
            self.fail(msg='The outpath was not valid html/xml!')

    def test_file_outpath_tiff(self):
        """
        Test that kraken creates hocr output for tiffs.
        """
        tiffpath = os.path.join(resources, u'image_tiff.tiff')
        modelpath = os.path.join(resources, u'en-default.pyrnn.gz')
        ocr = kraken.ocr(tiffpath, modelpath)
        try:
            parser = etree.HTMLParser()
            etree.fromstring(ocr, parser)
        except etree.XMLSyntaxError:
            self.fail(msg='The outpath was not valid html/xml!')

    def test_file_outpath_jpg(self):
        """
        Test that kraken creates hocr output for jpgs.
        """
        jpgpath = os.path.join(resources, u'image_jpg.jpg')
        modelpath = os.path.join(resources, u'en-default.pyrnn.gz')
        ocr = kraken.ocr(jpgpath, modelpath)
        try:
            parser = etree.HTMLParser()
            etree.fromstring(ocr, parser)
        except etree.XMLSyntaxError:
            self.fail(msg='The outpath was not valid html/xml!')

    def test_nlbin(self):
        """
        Test that kraken's nlbin is callable.
        """
        jpgpath = os.path.join(resources, u'image_jpg.jpg')
        outpath = os.path.join(self.tempdir, u'output.png')
        kraken.kraken_nlbin(jpgpath, outpath)
        self.assertTrue(os.path.isfile(outpath),
                        msg='Kraken did not output a file!')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
nidaba.plugins
==============

"""

from __future__ import absolute_import, print_function

import os

from pluginbase import PluginBase
from nidaba.config import nidaba_cfg


plugin_base = PluginBase(package='nidaba.plugins')
sp = [os.path.dirname(__file__)]
if 'plugin_path' in nidaba_cfg:
    sp.extend(nidaba_cfg['plugin_path'])
plugin_source = plugin_base.make_plugin_source(searchpath=sp)

with plugin_source:
    if 'plugins_load' in nidaba_cfg:
        for plugin in nidaba_cfg['plugins_load']:
            pl = plugin_source.load_plugin(plugin)
            pl.setup(**nidaba_cfg['plugins_load'][plugin])
    else:
        __import__('nidaba.plugins', globals(), locals(), plugin_source.list_plugins(), -1)



"""
nidaba.algorithms.otsu
~~~~~~~~~~~~~~~~~~~~~~

Module implementing variants of Otsu's method.

"""

import numpy as np


def otsu(im):
    """
    A naive native python implementation of Otsu thresholding (or at least the
    algorithm Wikipedia describes as Otsu's method).

    Args:
        im (PIL.Image): A PIL Image object in mode 'L' (8bpp grayscale)

    Returns:
        PIL.Image in mode '1' (1bpp b/w) containing the binarized image
    """

    assert im.mode == 'L'
    hist = im.histogram()
    total = np.prod(im.size)
    st = np.inner(range(0, len(hist)), hist)

    wb = 0.0
    sb = 0.0

    thresh = -1
    mvar = 0.0
    for i in range(0, len(hist)):
        wb += hist[i]
        if wb == 0:
            continue
        wf = total - wb
        if wf == 0:
            break
        sb += i * hist[i]
        mb = sb / wb
        mf = (st - sb)/wf
        bcv = wb * wf * (mb-mf)**2
        if bcv > mvar:
            mvar = bcv
            thresh = i

    return im.point(lambda p: p > thresh and 255, mode='1')



# -*- coding: utf-8 -*-
"""
nidaba.tasks.binarize
~~~~~~~~~~~~~~~~~~~~~

Tasks for segmentation of color images into binary images. More often than not
these will operate on grayscale input.
"""

from __future__ import absolute_import

from nidaba import storage
from nidaba import image
from nidaba.celery import app
from nidaba.tasks.helper import NidabaTask


@app.task(base=NidabaTask, name=u'nidaba.binarize.otsu')
def otsu(doc, method=u'otsu'):
    """
    Binarizes an input document utilizing a naive implementation of Otsu's
    thresholding.

    Args:
        doc (unicode, unicode): The input document tuple.
        method (unicode): The suffix string appended to all output files.

    Returns:
        (unicode, unicode): Storage tuple of the output file

    """
    input_path = storage.get_abs_path(*doc)
    output_path = storage.insert_suffix(input_path, method)
    return storage.get_storage_path(image.otsu(input_path, output_path))



# -*- coding: utf-8 -*-
import unittest
import os
import shutil
import tempfile
import subprocess
import ctypes

from lxml import etree
from nidaba.plugins import tesseract
from distutils import spawn
from nose.plugins.skip import SkipTest

thisfile = os.path.abspath(os.path.dirname(__file__))
tessdata = os.path.abspath(os.path.join(thisfile, 'resources'))


class TesseractTests(unittest.TestCase):

    """
    Tests the tesseract plugin.
    """

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_capi_multiple(self):
        """
        Test that tesseract CAPI calls create hocr output for multiple
        languages.
        """
        try:
            t = ctypes.cdll.LoadLibrary('libtesseract.so.3')
        except:
            raise SkipTest

        tiffpath = os.path.join(tessdata, 'image.tiff')
        outpath = os.path.join(self.tempdir, 'output')
        tesseract.setup(tessdata=tessdata, implementation='capi')
        tesseract.ocr_capi(tiffpath, outpath, ['grc', 'eng'])
        self.assertTrue(os.path.isfile(outpath),
                        msg='Tesseract did not output a file!')
        try:
            etree.parse(outpath)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')

    def test_direct_multiple(self):
        """
        Test that direct tesseract calls create hocr output for multiple
        languages.
        """
        if not spawn.find_executable('tesseract'):
            raise SkipTest

        tiffpath = os.path.join(tessdata, 'image.tiff')
        outpath = os.path.join(self.tempdir, 'output')
        tesseract.setup(tessdata=tessdata, implementation='direct')
        tesseract.ocr_direct(tiffpath, outpath, ['grc', 'eng'])
        if os.path.isfile(outpath + '.html'):
            outpath = outpath + '.html'
        else:
            outpath = outpath + '.hocr'
        self.assertTrue(os.path.isfile(outpath),
                        msg='Tesseract did not output a file!')
        try:
            etree.parse(outpath)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')

    def test_capi_file_output_png(self):
        """
        Test that tesseract CAPI calls create hocr output for pngs.
        """

        try:
            t = ctypes.cdll.LoadLibrary('libtesseract.so.3')
        except:
            raise SkipTest

        pngpath = os.path.join(tessdata, 'image.png')
        outpath = os.path.join(self.tempdir, 'output')
        tesseract.setup(tessdata=tessdata, implementation='capi')
        tesseract.ocr_capi(pngpath, outpath, ['grc'])
        self.assertTrue(os.path.isfile(outpath),
                        msg='Tesseract did not output a file!')
        try:
            etree.parse(outpath)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')

    def test_capi_file_output_tiff(self):
        """
        Test that tesseract CAPI calls create hocr output for tiffs.
        """
        try:
            t = ctypes.cdll.LoadLibrary('libtesseract.so.3')
        except:
            raise SkipTest

        tiffpath = os.path.join(tessdata, 'image.tiff')
        outpath = os.path.join(self.tempdir, 'output')
        tesseract.setup(tessdata=tessdata, implementation='capi')
        tesseract.ocr_capi(tiffpath, outpath, ['grc'])
        self.assertTrue(os.path.isfile(outpath),
                        msg='Tesseract did not output a file!')
        try:
            etree.parse(outpath)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')

    def test_capi_file_output_jpg(self):
        """
        Test that tesseract CAPI calls create hocr output for jpgs.
        """

        try:
            t = ctypes.cdll.LoadLibrary('libtesseract.so.3')
        except:
            raise SkipTest

        jpgpath = os.path.join(tessdata, 'image.jpg')
        outpath = os.path.join(self.tempdir, 'output')
        tesseract.setup(tessdata=tessdata, implementation='capi')
        tesseract.ocr_capi(jpgpath, outpath, ['grc'])
        self.assertTrue(os.path.isfile(outpath),
                        msg='Tesseract did not output a file!')
        try:
            etree.parse(outpath)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')

    def test_direct_file_output_png(self):
        """
        Test that direct tesseract calls create hocr output for pngs.
        """

        if not spawn.find_executable('tesseract'):
            raise SkipTest

        pngpath = os.path.join(tessdata, 'image.png')
        outpath = os.path.join(self.tempdir, 'output')
        tesseract.setup(tessdata=tessdata, implementation='direct')
        tesseract.ocr_direct(pngpath, outpath, ['grc'])
        if os.path.isfile(outpath + '.html'):
            outpath = outpath + '.html'
        else:
            outpath = outpath + '.hocr'
        self.assertTrue(os.path.isfile(outpath),
                        msg='Tesseract did not output a file!')
        try:
            etree.parse(outpath)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')

    def test_direct_file_output_tiff(self):
        """
        Test that direct tesseract calls create hocr output for tiffs.
        """
        if not spawn.find_executable('tesseract'):
            raise SkipTest

        tiffpath = os.path.join(tessdata, 'image.tiff')
        outpath = os.path.join(self.tempdir, 'output')
        tesseract.setup(tessdata=tessdata, implementation='direct')
        tesseract.ocr_direct(tiffpath, outpath, ['grc'])
        if os.path.isfile(outpath + '.html'):
            outpath = outpath + '.html'
        else:
            outpath = outpath + '.hocr'
        self.assertTrue(os.path.isfile(outpath),
                        msg='Tesseract did not output a file!')
        try:
            etree.parse(outpath)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')

    def test_direct_file_output_jpg(self):
        """
        Test that direct tesseract calls create hocr output for jpgs.
        """
        if not spawn.find_executable('tesseract'):
            raise SkipTest

        jpgpath = os.path.join(tessdata, 'image.jpg')
        outpath = os.path.join(self.tempdir, 'output')
        tesseract.setup(tessdata=tessdata, implementation='direct')
        tesseract.ocr_direct(jpgpath, outpath, ['grc'])
        if os.path.isfile(outpath + '.html'):
            outpath = outpath + '.html'
        else:
            outpath = outpath + '.hocr'
        self.assertTrue(os.path.isfile(outpath),
                        msg='Tesseract did not output a file!')
        try:
            etree.parse(outpath)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

from setuptools import setup

setup(
    setup_requires=['pbr', 'nose>=1.0', 'testrepository'],
    pbr=True,
    extras_require = {
        'kraken': ['kraken>=0.3.1']
    },
)




# -*- coding: utf-8 -*-

from __future__ import absolute_import

import yaml
import sys

from os import path


def reload_config():
    """
    Triggers a global reloading of the configuration files.

    As of now configuration files are only read from sys.prefix/etc/nidaba/.
    """
    global nidaba_cfg, celery_cfg
    ipath = path.join(sys.prefix, 'etc', 'nidaba', 'nidaba.yaml')
    with open(ipath, 'rb') as fp:
        nidaba_cfg = yaml.safe_load(fp)

    cpath = path.join(sys.prefix, 'etc', 'nidaba', 'celery.yaml')
    with open(cpath, 'rb') as fp:
        celery_cfg = yaml.safe_load(fp)

reload_config()



# -*- coding: utf-8 -*-
"""
nidaba.nidaba
~~~~~~~~~~~~~

The public API of nidaba. External applications should exclusively use the
objects and methods defined here.
"""

from __future__ import absolute_import

from nidaba import tasks
from nidaba import plugins
from nidaba import celery
from nidaba import storage
from nidaba.nidabaexceptions import (NidabaInputException,
                                     NidabaNoSuchAlgorithmException,
                                     NidabaTickException, NidabaStepException)

from itertools import product
from celery import chain
from celery import group
from celery.result import AsyncResult
from celery.states import state

import json


class Batch(object):

    """
    Creates a series of celery tasks OCRing a set of documents (among other
    things).

    A batch contains three level of definitions: tasks, ticks, and steps. A
    task is a singular operation on an input document, creating a single output
    document, e.g. binarization using a particular configuration of an
    algorithm or OCR using a particular engine. Multiple tasks executing in
    parallel are grouped into a tick and multiple ticks (running sequentially)
    are grouped into steps which again are executed sequentially.

    The need for steps and ticks arises from two different execution orders
    required by a pipeline. Take the following example:

        step 1:
            tick a: task 1
            tick b: task 2, task 3
            tick c: task 4, task 5

    The pipeline expands this example to the following sequences run in
    parallel (dot product of all ticks):

        task 1 -> task 2 -> task 4
        task 1 -> task 2 -> task 5
        task 1 -> task 3 -> task 4
        task 1 -> task 3 -> task 5

    It is not garantueed that any particular task in another sequence has
    executed successfully before a task in a sequence is run, i.e. it is not
    ensured that all task 1's have finished before task 2 of the first sequence
    is executed, except the task(s) further up the sequence.

    Steps on the other hand ensure that all tasks of the previous step have
    finished successfully. The output(s) of the expanded ticks is aggregated
    into a single list and used as the input of the first tick of the step.
    Expanding on the example the following step is added:

        step 2:
            tick d: task 6

    After the 4 sequence are finished their output is aggregated into a list
    [d1, d2, d3, d4] and used as the input of task 6. The final output of task
    6 is the output of the pipeline.

    The call order to create this example is:

        Batch.add_step()
        Batch.add_tick()
        Batch.add_task(task_1)
        Batch.add_tick()
        Batch.add_task(task_2)
        Batch.add_task(task_3)
        Batch.add_tick()
        Batch.add_task(task_4)
        Batch.add_task(task_5)
        Batch.add_step()
        Batch.add_tick()
        Batch.add_task(task_6)
    """

    def __init__(self, id):

        self.id = id
        self.docs = []
        self.batch_def = []
        self.cur_step = None
        self.cur_tick = None

    def get_state(self):
        """Retrieves the current state of a batch.

        Returns:
            (unicode): A string containing one of the following states:

                NONE: The batch ID is not registered in the backend.
                FAILURE: Batch execution has failed.
                PENDING: The batch is currently running.
                SUCCESS: The batch has completed successfully.
        """
        batch = celery.app.backend.get(self.id)
        try:
            batch = json.loads(batch)
        except Exception:
            return u'NONE'
        if len(batch['errors']) > 0:
            return u'FAILURE'

        st = state('SUCCESS')
        for id in batch['task_ids']:
            if AsyncResult(id).state < st:
                st = AsyncResult(id).state
        return unicode(st)

    def get_errors(self):
        """
        Retrieves all errors of the batch.

        Returns:
            list: A list of tuples containing keyword arguments to the task, a
            dictionary containing debug tracking information (i.e. variables
            which are given to the tasks as keyword arguments but aren't
            arguments to the underlying function), and the exception message of
            the failure.
        """
        batch = celery.app.backend.get(self.id)
        try:
            batch = json.loads(batch)
        except:
            return None
        if len(batch['errors']) > 0:
            return batch['errors']

        return None

    def get_results(self):
        """
        Retrieves the storage tuples of a successful batch or None.

        Returns:
            A list of dictionaries containing the output document name under
            the key 'doc', the original document it is based upon under the key
            'root', and all further tracking data that may be added during
            execution.
        """
        batch = celery.app.backend.get(self.id)
        try:
            batch = json.loads(batch)
        except Exception:
            return None

        if self.get_state() != 'SUCCESS':
            return None

        outfiles = []
        for id in batch['task_ids']:
            ch = AsyncResult(id)
            if ch.successful():
                if type(ch.result['doc'][0]) == dict:
                    outfiles.extend([x for x in ch.result['doc']])
                else:
                    outfiles.append(ch.result)
        return outfiles

    def add_document(self, doc):
        """Add a document to the batch.

        Adds a document tuple to the batch and checks if it exists.

        Args:
            doc (tuple): A standard document tuple.

        Raises:
            NidabaInputException: The document tuple does not refer to a file.
        """

        if not storage.is_file(*doc):
            raise NidabaInputException('Input document is not a file.')
        self.docs.append(doc)

    def add_task(self, method, **kwargs):
        """Add a task to the current tick.

        Adds a ``task``, a single executable task gathering one or more input
        documents and returning a single output document, to the current tick.
        Multiple jobs are run in parallel.

        Args:
            method (unicode): A task identifier
            **kwargs: Arguments to the task

        Raises:
            NidabaTickException: There is no tick to add a task to.
            NidabaNoSuchAlgorithmException: Invalid method given.
        """
        if self.cur_tick is None:
            raise NidabaTickException('No tick to add task to.')
        if u'nidaba.' + method not in celery.app.tasks:
            raise NidabaNoSuchAlgorithmException('No such task in registry')
        kwargs[u'method'] = method
        kwargs[u'id'] = self.id
        self.cur_tick.append(kwargs)

    def add_tick(self):
        """Add a new tick to the current step.

        Adds a ``tick``, a set of tasks running in parallel and sharing common
        input documents to the current step.

        Raises:
            NidabaStepException: There is no step to add a tick to.
        """
        if self.cur_step is None:
            raise NidabaStepException('No step to add tick to.')
        if self.cur_tick:
            self.cur_step.append(self.cur_tick)
        self.cur_tick = []

    def add_step(self):
        """Add a new step to the batch definition.

        Adds a ``step``, a list of sequentially run ticks to the batch
        definition. The output(s) of the last tick of a step is aggregated into
        a single list and used as the input of the first tick of the following
        step.
        """
        if self.cur_tick:
            self.cur_step.append(self.cur_tick)
            self.cur_tick = []
        if self.cur_step:
            self.batch_def.append(self.cur_step)
        self.cur_step = []

    def run(self):
        """Executes the current batch definition.

        Expands the current batch definition to a series of celery chords and
        executes them asynchronously. Additionally a batch record is written to
        the celery result backend.

        Returns:
            (unicode): Batch identifier.
        """
        # flush most recent tick/step onto the batch definition
        self.add_tick()
        self.add_step()

        groups = []
        tick = []
        # We first expand the tasks starting from the second step as these are
        # the same for each input document.
        if len(self.batch_def) > 1:
            groups.append(tasks.util.sync.s())
            for tset in self.batch_def[1:]:
                for sequence in product(*tset):
                    method = celery.app.tasks[
                        'nidaba.' + sequence[0]['method']]
                    ch = chain(method.s(**(sequence[0])))
                    for seq in sequence[1:]:
                        method = celery.app.tasks['nidaba.' + seq['method']]
                        ch |= method.s(**seq)
                    tick.append(ch)
                groups.append(group(tick))
                groups.append(tasks.util.sync.s())

        # The expansion steps described above is redone for each input document
        rets = []
        for doc in self.docs:
            tick = []
            # the first step is handled differently as the input document has
            # to be added explicitely.
            for sequence in product([doc], *self.batch_def[0]):
                method = celery.app.tasks['nidaba.' + sequence[1]['method']]
                ch = chain(method.s(doc=sequence[0], root=sequence[0],
                                    **(sequence[1])))
                for seq in sequence[2:]:
                    method = celery.app.tasks['nidaba.' + seq['method']]
                    ch |= method.s(**seq)
                tick.append(ch)
            rets.append(
                chain([group(tick)] + [tasks.util.sync.s()] +
                      groups[:-1]).apply_async().id)
        celery.app.backend.set(
            self.id, json.dumps({'errors': [], 'task_ids': rets}))
        return self.id



# -*- coding: utf-8 -*-
# This module contains all file handling/storage management/ID mapping methods

from __future__ import absolute_import

from nidaba.lock import lock
from nidaba.config import nidaba_cfg
from nidaba.nidabaexceptions import (NidabaStorageViolationException,
                                     NidabaNoSuchStorageBin)

import os
import fnmatch
import re


def _sanitize_path(base_path, *paths):
    """
    Sanitizes a given path with respect to a base path.

    Args:
        base_path (unicode): A base path beneath the concatenation of *paths is
        garantueed to be in.
        *paths (unicode): A list of subpaths which are concatenated and then
        checked if they are beneath base_path.

    Returns:
        (unicode): A string of the absolute path of the concatenations of
                   *paths.

    Raises:
        NidabaStorageViolationException: The absolute path of *paths is not
                                         beneath base_path. And whatever the
                                         python standard library decides to
                                         raise this week.
    """

    if len(paths) < 1:
        raise NidabaStorageViolationException('Path not beneath STORAGE_PATH')
    base_path = os.path.expanduser(base_path)
    base_path = os.path.abspath(base_path)
    rel_path = os.path.abspath(os.path.join(base_path, *paths))
    if os.path.commonprefix([os.path.normpath(rel_path),
                             os.path.normpath(base_path)]) == base_path:
        return rel_path
    else:
        raise NidabaStorageViolationException('Path not beneath STORAGE_PATH')


def is_file(jobID, path):
    """
    Checks if a storage tuple is a regular file.

    Args:
        jobID (unicode): An unique ID associated with a particular job.
        path (unicode): A path of a file beneath jobID.

    Returns:
        bool: Either True or False depending on the existence of the file.

    Raises:
        Exception: Who the fuck knows. The python standard library doesn't
                   document such paltry information as exceptions.
    """
    return os.path.isfile(get_abs_path(jobID, path))


def get_abs_path(jobID, *path):
    """
    Returns the absolute path of a file.

    Takes a job ID and a sequence of path components and checks if their
    absolute path is in the directory of that particular job ID.

    Args:
        jobID (unicode): A unique job ID
        *path (unicode): A list of path components that are concatenated to
        calculate the absolute path.

    Returns:
        (unicode): A string containing the absolute path of the storage tuple.

    Raises:
        NidabaStorageViolationException: The resulting absolute path is
                                         either not in the storage_path of the
                                         nidaba configuration or not in its job
                                         directory.
    """
    if len(path) < 1:
        raise NidabaStorageViolationException('Path not beneath STORAGE_PATH')
    # Run twice to ensure resulting path is beneath jobID.
    return _sanitize_path(_sanitize_path(nidaba_cfg['storage_path'], jobID),
                          *path)


def get_storage_path(path):
    """
    Converts an absolute path to a storage tuple of the form (id, path).

    Args:
        path (unicode): A unicode string of the absolute path.

    Returns:
        tuple: (id, path)

    Raises:
        NidabaStorageViolationException: The given path can not be converted
                                         into a storage tuple.
        NidabaNoSuchStorageBin: The given path is not beneath a valid job ID.
    """
    base_path = _sanitize_path(nidaba_cfg['storage_path'], u'')
    if os.path.commonprefix([os.path.normpath(base_path),
                             os.path.normpath(path)]) != base_path:
        raise NidabaStorageViolationException('Path not beneath STORAGE_PATH')
    path = path.replace(base_path, u'', 1)
    m = re.match('^(?P<id>.+?)\/(?P<p>.+)', path)
    id = os.path.split(m.groupdict()['id'])[1]
    if is_valid_job(id):
        return (id, m.groups()[1])
    else:
        raise NidabaNoSuchStorageBin('ID ' + m.groupdict()['id'] + ' not\
                                     known.')


def insert_suffix(orig_path, *suffix):
    """
    Inserts one or more suffixes just before the file extension.
    """
    pathname, extension = os.path.splitext(orig_path)
    for i in suffix:
        pathname += u'_' + i
    return pathname + extension


def is_valid_job(jobID):
    """
    Checks if filestore has been prepared for a job.

    Args:
        jobID (unicode): An identifier of a job.

    Returns:
        bool: True if job is already in the system, False otherwise.

    Raises:
        Standard python library caveats apply.
    """
    return os.path.isdir(_sanitize_path(nidaba_cfg['storage_path'], jobID))


def prepare_filestore(jobID):
    """
    Prepares the default filestore to accept files for a job.

    Args:
        jobID (unicode): Identifier of the bin to be created.

    Returns:
        None: Failure
        (unicode): String containing the job ID
    """
    if is_valid_job(jobID):
        return None
    try:
        jobPath = _sanitize_path(nidaba_cfg['storage_path'], jobID)
        os.mkdir(jobPath)
        return jobID
    except Exception:
        return None


def list_content(jobID, pattern=u'*'):
    """
    Lists all files to a job ID, optionally applying a glob-like filter.

    Args:
        jobID (unicode): Identifier of the bin
        pattern (unicode): glob-like filter to match files

    Returns:
        list: A list of unicode strings of the matching files.
    """
    if not is_valid_job(jobID):
        return None
    flist = []
    jpath = _sanitize_path(nidaba_cfg['storage_path'], jobID)
    for root, dirs, files in os.walk(jpath):
        flist.extend([os.path.relpath(os.path.join(root, s), jpath)
                      for s in files])
    return fnmatch.filter(flist, pattern)


def retrieve_content(jobID, documents=None):
    """
    Retrieves data from a single or a list of documents. Returns binary data,
    for retrieving unicode text use retrieve_text().

    Args:
        jobID (unicode): Identifier of the bin
        documents (tuple or list of tuples): Documents to read in

    Returns:
        None: Failure
        Dictionary: A dictionary mapping file identifiers to their contents.
    """
    if not is_valid_job(jobID):
        return None
    if documents:
        if isinstance(documents, basestring):
            documents = [documents]
        fdict = {}
        dpath = _sanitize_path(nidaba_cfg['storage_path'], jobID)
        locks = [lock(_sanitize_path(dpath, doc)) for doc in documents]
        map(lambda x: x.acquire(), locks)
        for doc in documents:
            with open(_sanitize_path(dpath, doc), 'rb') as f:
                fdict[doc] = f.read()
        map(lambda x: x.release(), locks)
        return fdict


def retrieve_text(jobID, documents=None):
    """
    Retrieves UTF-8 encoded text from a single or a list of documents.

    Args:
        jobID (unicode): Identifier of the bin
        documents (tuple or list of tuples): Documents to read in

    Returns:
        None: Failure
        Dictionary: A dictionary mapping file identifiers to their contents.
    """
    res = retrieve_content(jobID, documents)
    return {t: res[t].decode('utf-8') for t in res}


def write_content(jobID, dest, data):
    """
    Writes data to a document at a destination beneath a jobID. Writes bytes,
    does not accept unicode objects; use write_text() for that.

    Args:
        jobID (unicode): Identifier of the bin.
        dest (tuple): Documents to write to.
        data (str): Data to write.

    Returns:
        int: Length of data written
        None: Failure
    """
    if not is_valid_job(jobID):
        return None
    if not isinstance(data, basestring):
        return None
    try:
        with open(_sanitize_path(nidaba_cfg['storage_path'],
                                 os.path.join(jobID, dest)), 'wb') as f:
            l = lock(f.name)
            l.acquire()
            f.write(data)
            l.release()
    except:
        return None
    return len(data)


def write_text(jobID, dest, text):
    """
    Writes text data encoded as UTF-8 to a file beneath a jobID.

    Args:
        jobID (unicode): Identifier of the bin.
        dest (tuple): Documents to write to.
        text (unicode): Data to write.

    Returns:
        int: Length of data written
    """
    return write_content(jobID, dest, text.encode('utf-8'))
